try:
    import pymysql
    from pymysql.cursors import DictCursor
except:
    print('you need to install pymysql\n$ : python -m pip install pymysql')
import traceback

from typing import Iterable, Union

from sosin.databases.db import Database

class MariaDB(Database):
    """
    MariaDB
    """
    
    def __init__(self, db_config:dict, cursor_type:str = None) -> None:
        db_config['port'] = int(db_config.get('port', '3306'))
        if cursor_type == 'dict':
            self.cursor_type = DictCursor
        else:
            self.cursor_type = None

        super().__init__(db_config, cursor_type)

    def connect(self):
        self.DB = pymysql.connect(**self.db_config)
        self.cursor = self.DB.cursor(cursor=self.cursor_type)
        return self
    
    def create_table(self, tb_name:str, fields:list, foreign_keys:list=[], auto_pk=True):
        """
        create table
        
        *required
        tb_name = 'tableA'
        fields = [
            # nullable, auto_increment, is_key can be omissible
            # nullable ~ is_key : type = boolean
            ('LastName', 'VARCHAR(255)', nullable, auto_increment, is_key), 
            
        ]
        foreign_keys = [
            ('column_name', 'target_table', 'target_column')
        ]
        """
        assert (sum([field[4] for field in fields if len(field) == 5]) + auto_pk) == 1, 'len(pk) to be 1'

        columns = ['id INT PRIMARY KEY' if auto_pk else '']
        for field in fields:
            if len(field) > 2:
                field[2] = 'NULL' if field[2] else 'NOT NULL'
            if len(field) > 3:
                field[3] = 'AUTO_INCREMENT' if field[3] else ''
            if len(field) > 4:
                field[4] = 'PRIMARY KEY' if field[4] else ''
            columns.append(' '.join(field))
        
        constraints = []
        for fk in foreign_keys:
            fk_column = fk[0]
            fk_t_table = fk[1]
            fk_t_column = fk[2]
            constraints.append(f', FOREIGN KEY ({fk_column}) REFERENCES {fk_t_table} ({fk_t_column})')

        try:
            # foreign_keys -> sql 문 생성
            with self.DB.cursor() as cursor:
                sql = 'CREATE TABLE %s ('%tb_name + \
                ', '.join(columns) + \
                ''.join(constraints) + \
                ');'
                cursor.execute(sql)
                self.DB.commit()
                return True
        except:
            traceback.print_exc()
            self.DB.rollback()
            return False
    
    def drop_table(self, tables:Iterable[str], forcing=True) -> bool:
        """
        drop table

        *required
        tables = ('tableA', 'tableB')
        """

        try:
            with self.DB.cursor() as cursor:
                sql = 'DROP TABLE %s;'%(', '.join(tables))
                cursor.execute(sql)
                self.DB.commit()
                return True
        except:
            traceback.print_exc()
            self.DB.rollback()
            return False
    
    def get_tables(self) -> list:
        """
        get table names
        """
        sql_qr = 'show tables;'
        self.execute(sql_qr)
        result = self.cursor.fetchall()
        return [t[0] for t in result]
    
    def get_table_columns(self, table:str) -> list[dict]:
        """
        get table column infos
        """

        sql_qr = 'SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE table_name=%s'
        self.cursorcursor.execute(sql_qr, table)
        
        return [{'idx': field[4], 'name': field[3], 'type': field[7],
                    'null': field[6] != 'NO', 'PK': field[16]=='PRI', 'FK': field[16]=='MUL'
                    } for field in self.cursorcursor.fetchall()]
    
    def custom_select(self, qry):
        return super().select(qry)

    def select(self, column_qry:str, table:str, limit=None, offset=None, order_by=None, where_condition=[]) -> tuple:
        """
        Select
        
        example) 
        column_qry = "*"
        column_qry = "id, name, email"
        table = "Students"
        """
        sql_qr = "SELECT {0} FROM {1}".format(column_qry, table)
        if where_condition:
            for i, (col, eq, val) in enumerate(where_condition):
                is_equal = '=' if eq else '!='
                is_multiple = ' AND' if i > 1 else ' WHERE'
                sql_qr += f'{is_multiple} {col}{is_equal}{val}'
        if order_by:
            sql_qr += ' ORDER BY {}'.format(order_by)
        if limit:
            sql_qr += ' LIMIT {}'.format(limit)
        if offset:
            sql_qr += ' OFFSET {}'.format(offset)

        return super().select(sql_qr)
    
    def insert(self, table:str, columns: str, value: tuple) -> Union[int, bool]:
        """
        Insert Data
        
        example)
        table = "Students"
        columns = "name, email, phone"
        values = ('이름', '이메일', '번호')
        """

        sql_qr = f"INSERT INTO {table}({columns}) " \
                  "VALUES (" +','.join(["%s"]*len(value)) +")"
        
        result = super().insert(sql_qr, value)
        if result:
            return self.cursor.lastrowid
        return result
    
    # values는 list 형식으로 넣었음, args로 함
    def insert_many(self, table:str, columns: str, values: list, ignore:bool=False) -> bool:
        """
        Insert Many Datas
        
        example)
        table = "Students"
        columns = "name, email, phone"
        values = [
            ('hong gildong', 'hgd123@gmail.com', '01012345678'),
            ...
        ]
        """
        ignore_sql = 'IGNORE ' if ignore else ''
        sql = f"INSERT {ignore_sql} INTO {table}({columns}) " \
                  "VALUES ("  + ','.join(["%s"]*len(values[0])) + ");"
        
        return super().insert_many(sql, values)

    def update(self, table:str, set_columns:list[str], set_values:list[str], where_column:str, where_value) -> bool:
        """
        Update
        
        example)
        table = "Students"
        set_column = ["name"]
        set_value = ["jason"]
        where_column = "id"
        where_value = "1"
        """

        set_statement = ', '.join('{}="{}"'.format(sc, sv) for sc, sv in zip(set_columns, set_values))
        sql = "UPDATE {0} " \
            "SET {1} " \
            "WHERE {2}={3};".format(table, set_statement, where_column, where_value)

        return super().update(sql)
    
    def delete(self, table:str, where_column:str, where_value) -> bool:
        """
        Delete
        
        example)
        table = "Students"
        where_column = "id"
        where_value = "1"
        """
        sql = 'DELETE FROM {0} WHERE {1}={2}'.format(table, where_column, where_value)
        
        return super().delete(sql)
    
    def truncate(self, table:str, forcing=True) -> bool:
        """
        truncate table
        """
        try:
            self.execute(f'SET FOREIGN_KEY_CHECKS = {int(forcing)}; TRUNCATE TABLE {table};')
            self.commit()
            return True
        except:
            return False
        
    # Procedure
    # unavailable
    def make_procedure(self, sp_name:str, inputs:list[tuple], outputs: list[tuple], variables:list[tuple], queries:list[tuple],):
        """
        inputs: [(input_name, dtype)]
        outputs: [(output_name, dtype)]
        varaibles: [(var_name, dtype)]
        """
        return None
        # ASSERT variables IN inputs

        input_str = ', '.join([' '.join(ip) for ip in inputs])
        if not [out for out in outputs if out[0] == 'RESULT']:
            outputs += [('RESULT', 'INT')]
        output_str = (', ' + ', '.join(['OUT ' + ' '.join(out) for out in outputs])) if outputs else ''
        variable_str = '\n'.join('DECLARE {} {};'.format(*var) for var in variables)

        query_strs = []
        for query in queries:
            if query[0].lower() == 'select':
                # select query
                query_strs.append(f'SELECT {query[1]} INTO {query[4]} FROM {query[2]} WHERE {query[3]} LIMIT 1;')
            elif query[0].lower() == 'insert':
                # insert query
                query_strs.append(f'INSERT INTO {query[2]}({query[1]}) VALUES ({query[3]});')
        
        query_str = '\n'.join(query_strs)
        
        query = f"""DELIMITER $$
        CREATE PROCEDURE `{sp_name}`({input_str}{output_str})
        BEGIN
            {variable_str}
            DECLARE exit handler for SQLEXCEPTION
            BEGIN
                ROLLBACK;
                SET RESULT = 0;  
            END;
        
        START TRANSACTION;
            {query_str}
            COMMIT;
            SET RESULT = 1;
        END$$"""
        
        with self.DB.cursor(self.cursor_type) as cursor:
            # do not work
            cursor.execute(query)
            self.DB.commit()
        return query
    # sample
    # self.make_procedure('sp_insert_order2', 
    # inputs=[('_oc_id', 'INT'), ('_commerce_id', 'TINYINT'), ('_sub_id', 'VARCHAR(32)'), ('_order_date', 'DATETIME'), 
    # ('_progress', 'VARCHAR(32)'), ('_quantity', 'INT'), ('_total_amount', 'INT'), ('_cal_amount', 'INT'), ('_buyer_name', 'VARCHAR(32)'), 
    # ('_buyer_phone', 'VARCHAR(32)'), ('_receiver_name', 'VARCHAR(32)'), ('_receiver_phone', 'VARCHAR(32)'), ('_receiver_address', 'VARCHAR(32)'), 
    # ('_PCC', 'VARCHAR(32)'), ('_product_id', 'VARCHAR(32)')], outputs=[], variables=[('_sp_id', 'INT')], 
    # queries=[('select', 'id', 'StoreProduct', 'sub_id=_product_id', '_sp_id'), 
    # ('insert', 'oc_id,sp_id,commerce_id,sub_id,order_date,progress,quantity,total_amount,cal_amount,buyer_name,buyer_phone,\
    # receiver_name,receiver_phone,receiver_address,PCC', 'OrderDetail', '_oc_id, _sp_id, _commerce_id, _sub_id, _order_date,\
    # _progress, _quantity, _total_amount, _cal_amount,_buyer_name,_buyer_phone, _receiver_name, _receiver_phone, \
    # _receiver_address, _PCC')])
    
    def call_procedure(self, sp_name:str, inputs:list, outputs:list[str]=['RESULT']):
        output_str = (',' + ','.join(['@'+output for output in outputs])) if outputs else ''

        with self.DB.cursor(self.cursor_type) as cursor:
            result = cursor.executemany(f'CALL {sp_name}('+','.join(["%s"]*len(inputs[0])) + output_str +');', 
                                        [self.make_values(o) for o in inputs])
        return result

    @staticmethod
    def make_columns(data: dict) -> str:
        """
        dictionary keys to columns str
        """
        return ','.join(data.keys())
    
    @staticmethod
    def make_values(data: dict) -> tuple:
        """
        dictionary values to tuple
        """
        return tuple(data.values())

if __name__ == '__main__':
    my_db = MariaDB({'host': 'localhost', 'port': 3306, 'user':'root', 'password': 'ppww', 'database': 'test'})

    with my_db:
        my_db.get_tables()
        
        fields = [
            ('name', 'VARCHAR(255)'),
            ('address', 'VARCHAR(255)'),
            ('phone', 'VARCHAR(32)'),
        ]
        my_db.create_table('tb_student', fields)

        fields = [
            ('student_id', 'INT'),
            ('math', 'INT'),
            ('science', 'INT'),
            ('english', 'INT'),
        ]
        foreign_keys = [
            ('student_id', 'tb_student', 'id')
        ]
        my_db.create_table('tb_grade', fields, foreign_keys)

        grade_columns = my_db.get_table_columns('tb_grade')
        print(grade_columns)

        my_db.drop_table(['tb_student', 'tb_grade'])
        ...



import time
try:
    from sosin.web.virtual import VirtualDriver, By, WebDriverWait, EC
except:
    pass

def papago(txts:list[str]) -> list[str]:
    
    """
    파파고 번역기
    """
    webdriver = VirtualDriver()
    webdriver.set_argument()
    driver = webdriver.get_driver()
    driver.execute_script("window.open('https://papago.naver.com/')")
    time.sleep(1)
    driver.switch_to.window(driver.window_handles[-1])
    result = []
    for txt in txts:
        WebDriverWait(driver,5).until(EC.presence_of_element_located((By.ID, 'txtSource'))).send_keys(txt)
        target_text = ''
        while True:
            # 충분히 번역할때까지 재확인
            time.sleep(2)
            if target_text == driver.find_element(By.ID, 'txtTarget').text:
                result.append(target_text)
                driver.find_element(By.ID, 'txtSource').clear()
                break
            target_text = driver.find_element(By.ID, 'txtTarget').text
        
    driver.quit()
    return result


# ---------------------------------------------------------------------------------------------
# Selnium Setting
try:
    from selenium import webdriver
    from selenium.webdriver.chrome.service import Service
    from webdriver_manager.chrome import ChromeDriverManager
    from webdriver_manager.core.os_manager import ChromeType
except:
    print('you need to install selenium, webdriver-manager\n$ : python -m pip install selenium webdriver-manager')

class VirtualDriver:
    """
    Selenium VirtualDriver
    """
    def __init__(self, engine='chrome') -> None:

        self.engine = engine.lower().strip()
        if self.engine  == 'chrome':
            self.options = webdriver.ChromeOptions()
        
    def set_argument(self, headless=True, linux=True, maximize=True, 
                        user_agent=False, lang_kr=False, secret_mode=False, download_path=None):
        """
        셀레니움 드라이버 옵션 설정
        headless: GUI 사용 안함
        gui: GUI 사용안할 시 추가
        maximize: 브라우저 최대화
        lang_kr: 한국어 설정
        secret_mode: 시크릿 모드
        download_path: headless 시 파일 다운로드 경로 설정
        """

        # 시스템에 부착된 ... 로그 제거
        self.options.add_experimental_option('excludeSwitches', ['enable-logging'])

        if headless:
            self.options.add_argument('--headless')

            if linux:
                # Bypass OS security model
                self.options.add_argument('--no-sandbox')
                self.options.add_argument("--disable-gpu")
                # overcome limited resource problems. 메모리가 부족해서 에러가 발생하는 것 막음
                self.options.add_argument('--disable-dev-shm-usage')
                
                # 크롬 드라이버에 setuid를 하지 않음으로써 크롬의 충돌 막음
                self.options.add_argument('--disable-setuid-sandbox')

                # disabling extensions
                self.options.add_argument("--disable-extensions")

                # 일단 좀 더 확인 필요
                self.options.add_argument('--single-process')

            if download_path:
                p = {
                "download.default_directory": download_path,
                "download.prompt_for_download": False,
                "download.directory_upgrade": True,
                "safebrowsing.enabled": True
                }
                self.options.add_experimental_option("prefs", p)

        if maximize:
            self.options.add_argument("--start-maximized")
        if user_agent:
            self.options.add_argument("user-agent=Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36")
        if lang_kr:
            self.options.add_argument("--lang=ko_KR")
        if secret_mode:
            self.options.add_argument("--incognito")

    def get_driver(self, is_chromium=False):
        """
        셀레니움 웹드라이버 실행
        """
        if self.engine == 'chrome':
            service = Service(
                ChromeDriverManager(chrome_type=ChromeType.CHROMIUM if is_chromium else ChromeType.GOOGLE).install())
            service.path = service.path.replace('THIRD_PARTY_NOTICES.chromedriver', 'chromedriver')
            driver = webdriver.Chrome(service=service, options=self.options)
            driver.maximize_window()
            driver.implicitly_wait(5)
            return driver


# ---------------------------------------------------------------------------------------------
# Selenium Utils
try:
    from selenium.webdriver.common.by import By
    from selenium.webdriver.common.keys import Keys
    from selenium.webdriver.common.alert import Alert
    from selenium.webdriver.support.ui import WebDriverWait
    from selenium.webdriver.support import expected_conditions as EC
except: pass

import time

MAX_RETRY_COUNT = 5
def click_alert(d):
    """
    alert
    """
    for _ in range(MAX_RETRY_COUNT):
        try:
            Alert(d).accept()
            break
        except: time.sleep(1)
    else: assert False, 'fail, alert not found'

if __name__ == '__main__':
    # 1. try-except
    try:
        driver = VirtualDriver().get_driver()
    except Exception as e:
        print(e)
    finally:
        driver.quit()

    # 2. class
    class MyWeb:
        def __init__(self):
            self.driver = VirtualDriver().get_driver()
            
        def __del__(self):
            self.driver.quit()



import random
import string

from typing import BinaryIO, Union

import httpx
from requests_toolbelt import MultipartEncoder

from .cookie import CookieManager
from sosin.utils.history import HistoryManager

class AsyncSessionManager(CookieManager):
    """
    Session Manager
    """

    history:list[httpx.Response] = []

    def __init__(self, cookie_path:str=None, history_mode:bool=False) -> None:
        self._headers = { 'User-Agent': 'Mozilla/5.0' }
        super().__init__(cookie_path)
        self.history_manager = HistoryManager() if history_mode else None
    
    # ------------------------------------------------------------------------
    # Request Management
    async def get(self, 
                  url:str, 
                  headers:dict={}, 
                  cookies:dict={}, 
                  params:dict={}, 
                  **kwargs
                  ) -> httpx.Response:
        
        r = await self._request(url, headers=headers, cookies=cookies, params=params,
                                method='get', **kwargs)
        return r

    async def post(self, 
                   url:str, 
                   headers:dict={}, 
                   cookies:dict={}, 
                   params:dict={},
                   data:Union[str, dict]={}, 
                   json:Union[dict, list]=None, 
                   files:dict[str]={}, 
                   **kwargs
                   ) -> httpx.Response:
        
        r = await self._request(url, headers=headers, cookies=cookies, params=params, 
                                data=data, files=files, json=json,
                                method='post', **kwargs)
        return r
    
    async def put(self, 
                  url:str, 
                  headers:dict={}, 
                  cookies:dict={}, 
                  params:dict={},
                  data:Union[str, dict, list]={}, 
                  json:Union[dict, list]=None, 
                  files:dict[str]={}, **kwargs
                  ) -> httpx.Response:
        
        r = await self._request(url, 
                                headers=headers, cookies=cookies, params=params, 
                                data=data, json=json, files=files, 
                                method='put', **kwargs)
        return r

    async def patch(self, 
                    url:str, 
                    headers:dict={}, 
                    cookies:dict={}, 
                    params:dict={},
                    data:Union[str, dict, list]={}, 
                    json:Union[list, dict]=None, 
                    files:dict[str]={}, 
                    **kwargs
                    ) -> httpx.Response:
        
        r = await self._request(url, 
                                headers=headers, cookies=cookies, params=params, 
                                data=data, json=json, files=files, 
                                method='patch', **kwargs)
        return r
    async def delete(self, 
                     url:str, 
                     headers:dict={}, 
                     cookies:dict={}, 
                     params:dict={},
                     data:Union[str, dict, list]={}, 
                     json:Union[list, dict]=None, 
                     files:dict[str]={}, 
                     **kwargs
                     ) -> httpx.Response:
        
        r = await self._request(url, 
                                headers=headers, cookies=cookies, params=params, 
                                data=data, json=json, files=files, 
                                method='delete', **kwargs)
        return r

    async def _request(self, 
                       url:str, 
                       headers:dict={}, 
                       cookies:dict={}, 
                       params:dict={}, 
                       data:Union[str, dict]={}, 
                       json:Union[dict, list]=None, 
                       files:dict[str]={}, 
                       method:str='post', 
                       **kwargs
                       ) -> httpx.Response:
        """
        files -> {key: file_path}
        """
        
        headers = {k.lower(): headers[k] for k in headers}

        type_header = {}
        if data:
            type_header['content-type'] = 'application/x-www-form-urlencoded'
        if json:
            type_header['content-type'] = 'application/json;charset=UTF-8'
        if files:
            boundary = '----WebKitFormBoundary' +\
                ''.join(random.sample(string.ascii_letters + string.digits, 16))
            data = MultipartEncoder(fields={**data, **{k: self.get_file_form(v) for k, v in files.items()}}, boundary=boundary)
            type_header['content-type'] = data.content_type
            type_header['connection'] = 'keep-alive'

        async with httpx.AsyncClient(verify=kwargs.pop('verify', True)) as client:
            if method.lower() == 'get':
                r = await client.get(url, 
                                     headers={**self._headers, **type_header, **headers}, 
                                     cookies={**self._cookies, **cookies}, 
                                     params=params, **kwargs)
            elif method.lower() == 'post':
                r = await client.post(url, 
                                      headers={**self._headers, **type_header, **headers},
                                      cookies={**self._cookies, **cookies},
                                      params=params, data=data, json=json, **kwargs)
            elif method.lower() == 'put':
                r = await client.put(url, 
                                     headers={**self._headers, **type_header, **headers},
                                     cookies={**self._cookies, **cookies}, 
                                     params=params, data=data, json=json, **kwargs)
            elif method.lower() == 'patch':
                r = await client.patch(url, 
                                       headers={**self._headers, **type_header, **headers}, 
                                       cookies={**self._cookies, **cookies}, 
                                       params=params, data=data, json=json, **kwargs)
            elif method.lower() == 'delete':
                r = await client.delete(url, 
                                        headers={**self._headers, **type_header, **headers}, 
                                        cookies={**self._cookies, **cookies}, 
                                        params=params, data=data, json=json, **kwargs)
            
        self._set_cookies(r)
        self._add_history(r)

        return r
    
    # ------------------------------------------------------------------------
    # History Functions
    def _add_history(self, r):
        if self.history_manager:
            self.history_manager.add_history(r)
    
    def get_histories(self) -> list:
        return self.history_manager.get_histories() if self.history_manager else []
    
    # ------------------------------------------------------------------------
    # Utils
    @staticmethod
    def get_file_form(f:str) -> tuple[str, BinaryIO]:
        """
        change file path to requestable formats
        """
        fn = f.split('/').pop()
        ext = fn.split('.').pop()
        mime_type = ''
        if ext in ['jpg', 'jpeg']:
            mime_type = 'image/jpeg'
        elif ext == 'png':
            mime_type = 'image/png'
        elif ext == 'xlsx':
            mime_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        elif ext == 'xls':
            mime_type = 'application/vnd.ms-excel'
        elif ext == 'zip':
            mime_type = 'application/x-zip-compressed'

        return fn, open(f, 'rb'), mime_type

if __name__ == '__main__':
    s = AsyncSessionManager()
    # s.get()
    # s.post()
    # s.put()



from datetime import datetime

def logging(t, log_path='./debug.log'):
    """
    log를 생성합니다.
    default path = ./debug.log
    """
    dt_now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    with open(log_path, 'a', encoding='UTF-8') as f:
        f.write(f'{dt_now} : {t}\n')



import requests

headers = {
    'user-agent': 'Mozilla/5.0'
}

class WebContentManager:
    """
    Web Content Manager
    download(path)
    get_content_type() -> image/jpeg, image/png, ...
    get_body() -> binary content
    """
    def __init__(self, url, method:str='get', add_headers:dict={}, add_body:dict={}) -> None:
        if method.lower()=='get':
            r = requests.get(url, headers={**headers, **add_headers})
        else:
            r = requests.post(url, headers={**headers, **add_headers}, body=add_body)
        r.raise_for_status()
        self.r = r

    def download(self, file_path:str):
        with open(file_path, 'wb') as f:
            f.write(self.get_body())
    
    def get_content_type(self):
        return self.r.headers['content-type']
    
    def get_body(self):
        return self.r.content

if __name__ == '__main__':
    wcm = WebContentManager('https://media.classicmanager.com/album/aac/MM014-2/01.m4a')

    wcm.download('./music.m4a')

    wcm.get_content_type()



# interface

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, overload
try:
    from pymysql.connections import Connection
    from pymysql.cursors import Cursor, DictCursor
except:
    ...

try:
    from psycopg2._psycopg import connection, cursor
    from psycopg2.extras import RealDictCursor
except:
    ...

try:
    class ConnectionType(Connection, connection): ...
    class CursorType(DictCursor, RealDictCursor): ...
    class CursorKindType(Cursor, cursor): ...
except:
    class ConnectionType(): ...
    class CursorType(): ...
    class CursorKindType(): ...

class Database(ABC):
    """
    데이터베이스 추상클래스
    """

    DB: ConnectionType
    db_config: Dict[str, str]
    cursor: Optional[CursorType] = None
    cursor_type: Optional[CursorKindType]
    
    def __init__(self, db_config: Dict[str, str], cursor_type:Optional[CursorKindType]=None) -> None:
        """
        **db_config**
            host=database host (localhost)
            port=port (3306)
            user=username (root)
            password=password (1q2w3e)
            database=database name (testdb)
            charset=charcter encoding (utf8mb4)
        """
        self.db_config = db_config
        self.cursor_type = cursor_type

    def __enter__(self):
        """
        Usage:
        ```py
        db = Database(config)
        with db: # as db_object
            ...
        ```
        """
        return self.connect()

    def __del__(self) -> None:
        try:
            self.disconnect()
        except:
            ...

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.disconnect()

    @abstractmethod
    def connect(self): ...

    def disconnect(self):
        if self.cursor:
            self.cursor.close()
        self.DB.close()

    def execute(self, query, args=None):
        self.cursor.execute(query, args)

    def executemany(self, query, args=None):
        self.cursor.executemany(query, args)

    def commit(self):
        self.DB.commit()
    
    @overload
    @abstractmethod
    def insert(self, 
               table: str, 
               columns: Union[str, List[str]], 
               value: Tuple
               ) -> bool: ...
    
    def insert(self, query, args=None) -> bool:
        try:
            self.execute(query, args)
            self.commit()
            return True
        except Exception as e:
            self.DB.rollback()
            print("INSERT ERR", e)
            return False

    @abstractmethod
    def insert_many(self, 
                    table: str, 
                    columns: Union[str, List[str]], 
                    values:List[Tuple]
                    ) -> bool: ...

    def insert_many(self, query, values) -> bool:
        try:
            self.executemany(query, values)
            self.commit()
            return True
        except Exception as e:
            self.DB.rollback()
            print("INSERT MANY ERR", e)
            return False

    @overload
    @abstractmethod
    def select(
        table: str, 
        columns: Union[str, Iterable], 
        where_info: Optional[Tuple[str, Any]]=None, 
        order_by: str=None, 
        page_info: Optional[Tuple[int, int]]=None): ...

    # not overloaded use just last function
    def select(self, query: str):
        try:
            self.cursor.execute(query)
            return self.cursor.fetchall()
        except Exception as e:
            return ("SELECT ERR", e)
    
    @abstractmethod
    def where(self, column: str, value: str, option: str=None): ...
    @abstractmethod
    def sort(self, order_by: str): ...
    @abstractmethod
    def pagination(self, page: int, size: int): ...

    @overload
    @abstractmethod
    def update(self, 
               table: str, 
               columns: List[str], 
               values: List, 
               where_info: Tuple[str, Any]
               ) -> bool: ...

    def update(self, query, args=None) -> bool:
        try:
            self.execute(query, args)
            self.commit()
            return True
        except Exception as e:
            self.DB.rollback()
            print("UPDATE ERR", e)
            return False

    @overload
    @abstractmethod
    def delete(self,
               table: str,
               where_info: Tuple[str, Any]
               ) -> bool: ...

    def delete(self, query, args=None) -> bool:
        try:
            self.execute(query, args)
            self.commit()
            return True
        except Exception as e:
            self.DB.rollback()
            print("DELETE ERR", e)
            return False



import random
import string

from typing import BinaryIO, Union

import requests
from requests_toolbelt import MultipartEncoder

from .cookie import CookieManager
from sosin.utils.history import HistoryManager

class SessionManager(CookieManager):
    """
    Session Manager
    """

    def __init__(self, 
                 cookie_path:str=None, 
                 history_mode:bool=False
                 ) -> None:
        super().__init__(cookie_path)
        self._headers = { 'User-Agent': 'Mozilla/5.0' }
        self.history_manager = HistoryManager() if history_mode else None
    
    # ------------------------------------------------------------------------
    # Request Management
    def get(self, 
            url:str, 
            headers:dict={}, 
            cookies:dict={},
            params:dict={}, 
            **kwargs
            ) -> requests.Response:
        
        r = self._request(url, 
                          headers=headers, cookies=cookies, params=params, 
                          method='get', **kwargs)
        return r

    def post(self, 
             url:str, 
             headers:dict={}, 
             cookies:dict={}, 
             params:dict={},
             data:Union[str, dict]={}, 
             json:Union[dict, list]=None, 
             files:dict[str]={}, 
             **kwargs
             ) -> requests.Response:
        
        r = self._request(url, 
                          headers=headers, cookies=cookies, params=params, 
                          data=data, json=json, files=files, 
                          method='post', **kwargs)
        return r
    
    def put(self, 
            url:str, 
            headers:dict={}, 
            cookies:dict={}, 
            params:dict={},
            data:Union[str, dict, list]={}, 
            json:Union[dict, list]=None, 
            files:dict[str]={}, 
            **kwargs
            ) -> requests.Response:
        
        r = self._request(url, 
                          headers=headers, cookies=cookies, params=params,
                          data=data, json=json, files=files, 
                          method='put', **kwargs)
        return r

    def patch(self, 
              url:str, 
              headers:dict={}, 
              cookies:dict={}, 
              params:dict={},
              data:Union[str, dict, list]={}, 
              json:Union[dict, list]=None, 
              files:dict[str]={}, 
              **kwargs
              ) -> requests.Response:
        
        r = self._request(url, 
                          headers=headers, cookies=cookies, params=params,
                          data=data, json=json, files=files, 
                          method='patch', **kwargs)
        return r
    
    def delete(self, 
               url:str, 
               headers:dict={}, 
               cookies:dict={}, 
               params:dict={},
               data:Union[str, dict, list]={}, 
               json:Union[dict, list]=None, 
               files:dict[str]={}, 
               **kwargs
               ) -> requests.Response:
        
        r = self._request(url, 
                          headers=headers, cookies=cookies, params=params,
                          data=data, json=json, files=files, 
                          method='delete', **kwargs)
        return r

    def _request(self, 
                 url:str, 
                 headers:dict={}, 
                 cookies:dict={}, 
                 params:dict={},
                 data:Union[str, dict]={}, 
                 json:Union[dict, list]={}, 
                 files:dict[str]={}, 
                 method:str='post', 
                 **kwargs
                 ) -> requests.Response:
        """
        files -> {key: file_path}
        """
        
        headers = {k.lower(): headers[k] for k in headers}

        type_header = {}
        if data:
            type_header['content-type'] = 'application/x-www-form-urlencoded'
        if json:
            type_header['content-type'] = 'application/json;charset=UTF-8'
        if files:
            boundary = '----WebKitFormBoundary' +\
                ''.join(random.sample(string.ascii_letters + string.digits, 16))
            data = MultipartEncoder(fields={**data, **{k: self.get_file_form(v) for k, v in files.items()}}, boundary=boundary)
            type_header['content-type'] = data.content_type
            type_header['connection'] = 'keep-alive'

        if method.lower() == 'get':
            r = requests.get(url, 
                             headers={**self._headers, **type_header, **headers},
                             cookies={**self._cookies, **cookies}, 
                             params=params, **kwargs)
        elif method.lower() == 'post':
            r = requests.post(url, 
                              headers={**self._headers, **type_header, **headers}, 
                              cookies={**self._cookies, **cookies},
                              params=params, data=data, json=json, **kwargs)
        elif method.lower() == 'put':
            r = requests.put(url, 
                             headers={**self._headers, **type_header, **headers}, 
                             cookies={**self._cookies, **cookies},
                             params=params, data=data, json=json, **kwargs)
        elif method.lower() == 'patch':
            r = requests.patch(url, 
                             headers={**self._headers, **type_header, **headers}, 
                             cookies={**self._cookies, **cookies},
                             params=params, data=data, json=json, **kwargs)
        elif method.lower() == 'delete':
            r = requests.delete(url, 
                             headers={**self._headers, **type_header, **headers}, 
                             cookies={**self._cookies, **cookies},
                             params=params, data=data, json=json, **kwargs)

        self._set_cookies(r)
        self._add_history(r)

        return r
    
    # ------------------------------------------------------------------------
    # History Functions
    def _add_history(self, r):
        if self.history_manager:
            self.history_manager.add_history(r)
    
    def get_histories(self) -> list:
        return self.history_manager.get_histories() if self.history_manager else []
    
    # ------------------------------------------------------------------------
    # Utils
    @staticmethod
    def get_file_form(f:str) -> tuple[str, BinaryIO]:
        """
        change file path to requestable formats
        """
        fn = f.split('/').pop()
        ext = fn.split('.').pop()
        mime_type = ''
        if ext in ['jpg', 'jpeg']:
            mime_type = 'image/jpeg'
        elif ext == 'png':
            mime_type = 'image/png'
        elif ext == 'xlsx':
            mime_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        elif ext == 'xls':
            mime_type = 'application/vnd.ms-excel'
        elif ext == 'zip':
            mime_type = 'application/x-zip-compressed'

        return fn, open(f, 'rb'), mime_type

if __name__ == '__main__':
    s = SessionManager()
    # s.get()
    # s.post()
    # s.put()



import setuptools

setuptools.setup(
    name                                = "sosin",
    version                             = "1.5.2",
    license                             = 'MIT',
    author                              = "Jason Choi",
    author_email                        = "sosincomp@gmail.com",
    description                         = "Python utils for general works",
    long_description                    = open('README.md').read(),
    url                                 = "https://github.com/devsosin/sosin",
    install_requires                    = ['requests', 'requests-toolbelt', 'httpx'],
    packages                            = setuptools.find_namespace_packages(include=['*']),
    python_requires                     = '>=3.9',
    classifiers                         = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ]
)